use anyhow::{Context, Result};
use app::App;
use modules::{config::ConfigModule, graphics::RenderModule};
use tokio::runtime::Builder;
use vulkano_util::{
    context::{VulkanoConfig, VulkanoContext},
    renderer::VulkanoWindowRenderer,
    window::{VulkanoWindows, WindowDescriptor},
};
use winit::{
    event::{Event, WindowEvent},
    event_loop::{ControlFlow, EventLoopBuilder},
};

mod app;
mod modules;
mod core;

fn main() -> Result<()> {
    let event_loop = EventLoopBuilder::new().build()?;
    
    let context = VulkanoContext::new(VulkanoConfig::default());
    let mut windows = VulkanoWindows::default();

    let runtime = Builder::new_multi_thread().enable_all().build()?;
    let (event_tx, event_rx) = flume::unbounded();

    runtime.block_on(async {
        runtime.spawn(async move {
            loop {
                let _event = event_rx.recv_async().await.unwrap();
                // println!(
                //     "Tokio got event: {:?} on thread: {:?}",
                //     event,
                //     std::thread::current().id()
                // );
            }
        });
    });

    let _id = windows.create_window(
        &event_loop,
        &context,
        &WindowDescriptor {
            title: "Khors".into(),
            present_mode: vulkano::swapchain::PresentMode::Fifo,
            ..Default::default()
        },
        |_| {},
    );

    let primary_window_renderer = windows.get_primary_renderer_mut().context("Failed to create primary window renderer")?;
    let _gfx_queue = context.graphics_queue();

    let mut app = App::new();    
    app.push_module(ConfigModule::new);
    app.push_module(RenderModule::new);

    event_loop
        .run(move |event, elwt| {
            elwt.set_control_flow(ControlFlow::Poll);

            if process_event(primary_window_renderer, &event, &mut app).expect("App execution failed") {
                elwt.exit();
            }

            event_tx.send(event.clone()).unwrap();
        })?;

    Ok(())
}

pub fn process_event(
    renderer: &mut VulkanoWindowRenderer,
    event: &Event<()>,
    app: &mut App,
) -> Result<bool> {
    match &event {
        Event::WindowEvent {
            event: WindowEvent::CloseRequested,
            ..
        } => {
            return Ok(true);
        }
        Event::WindowEvent {
            event: WindowEvent::Resized(..) | WindowEvent::ScaleFactorChanged { .. },
            ..
        } => renderer.resize(),
        Event::WindowEvent {
            event: WindowEvent::RedrawRequested,
            ..
        } => 'redraw: {
            // Tasks for redrawing:
            // 1. Update state based on events
            // 2. Compute & Render
            // 3. Reset input state
            // 4. Update time & title

            // The rendering part goes here:
            match renderer.window_size() {
                [w, h] => {
                    // Skip this frame when minimized.
                    if w == 0.0 || h == 0.0 {
                        break 'redraw;
                    }
                }
            }
            
            app.run()?;
        }
        Event::AboutToWait => renderer.window().request_redraw(),
        _ => (),
    }
    Ok(false)
}
